#include <oxstd.h>
#include <oxdraw.h>
#include <oxprob.h>
#include <oxfloat.h>
#include <arma.h>
#import <modelbase>
#import <simula>
#import <maximize>
#import <database>
#import <maxsqp>

enum{EXP, WEI, GAM, GEN_GAM, LOG_NORM, LAMBDA, LOG_LAMBDA, INV_FISHER, INV_SQRT_FISHER, HESS, SAND};

static decl s_mY, s_cT, s_vf, s_vScore, s_vScaledSc, s_iSandw, s_vFunc;
static decl s_iLinkFunction, s_iScalingChoice, s_iDistribution, s_ip, s_iq, s_cPar, s_iStdErr;
			   
StartingValues(dOmega, const vA, const vB, const dAlpha, const dKappa)
{
	decl asA, asB, vpar, saparnames;
	if(s_iLinkFunction == LAMBDA){
		if(dOmega <= 0) dOmega = -10;
		else dOmega = log(dOmega);
	}
	if(rows(vA) != s_ip || rows(vB) != s_iq){
		println("Number of starting values in vector vA, vB must be equal to s_ip, s_iq \nProgram stopped");
		exit(0);}
	if(sumc(vB) > 1){
		println("Sum of elements in vector vB cannot exceed 1 \nProgram stopped");
		exit(0);}
	asA = new array[1][s_ip];
	asB = new array[1][s_iq];
	for(decl i = 0; i < s_ip; i++)
		asA[0][i] = sprint("A", i+1);
	for(decl i = 0; i < s_iq; i++)
		asB[0][i] = sprint("B", i+1);
	vpar = dOmega | vA | vB | dAlpha | dKappa;
	saparnames = {"omega"} ~ asA[0][] ~ asB[0][] ~ {"alpha","kappa"};	
	if(s_iDistribution == EXP)	
		return {vpar[:s_ip+s_iq], saparnames[:s_ip+s_iq]};
	else if(s_iDistribution == WEI || s_iDistribution == GAM || s_iDistribution == LOG_NORM)	
		return {vpar[:s_ip+s_iq+1], saparnames[:s_ip+s_iq+1]};
	else if(s_iDistribution == GEN_GAM)	
		return {vpar, saparnames};		
}

SystemMatrices(const dY, const dLambda_t, const dAlpha, dKappa, const avScore, const amInvFisher, const adllik)
{
	if(s_iDistribution == EXP){
		avScore[0] = dY/sqr(dLambda_t) - 1/dLambda_t;
		amInvFisher[0] = sqr(dLambda_t);
		adllik[0] = -(dY/dLambda_t) - log(dLambda_t);
	} else if(s_iDistribution == WEI){
		avScore[0] = (dAlpha/dLambda_t)*((gammafact(1+(1/dAlpha))*(dY/dLambda_t))^dAlpha) - (dAlpha/dLambda_t);
		amInvFisher[0] = sqr(dLambda_t)/sqr(dAlpha);
		adllik[0] = log(dAlpha/dY) + dAlpha*log(dY/dLambda_t) + dAlpha*log(gammafact(1+(1/dAlpha))) - (gammafact(1+(1/dAlpha))*(dY/dLambda_t))^dAlpha;
	} else if(s_iDistribution == GAM){
		avScore[0] = ((dY*dAlpha)/sqr(dLambda_t)) - (dAlpha/dLambda_t);
		amInvFisher[0] = sqr(dLambda_t)/dAlpha;
		adllik[0] = dAlpha*log(dY*dAlpha/dLambda_t)  - log(dY*gammafact(dAlpha)) - (dY*dAlpha/dLambda_t);		
	} else if(s_iDistribution == GEN_GAM){
		decl dlambda_par = exp(loggamma(dKappa)-loggamma(dKappa+(1/dAlpha)));
		avScore[0] = (dAlpha/dLambda_t)*((dY/(dLambda_t*dlambda_par))^dAlpha) - ((dKappa*dAlpha)/dLambda_t);
		amInvFisher[0] = sqr(dLambda_t)/(dKappa*sqr(dAlpha));
		adllik[0] = log(dAlpha) - loggamma(dKappa) + (dKappa*dAlpha-1)*log(dY) - dKappa*dAlpha*log(dlambda_par*dLambda_t) - (dY/(dLambda_t*dlambda_par))^dAlpha;	
	} else if(s_iDistribution == LOG_NORM){
		avScore[0] = (log(dY)-log(dLambda_t)+0.5*dAlpha)/(dAlpha*dLambda_t);
		amInvFisher[0] = dAlpha*sqr(dLambda_t);
		adllik[0] = -log(dY) - 0.5*log(M_2PI) - 0.5*log(dAlpha) - sqr(log(dY)-log(dLambda_t)+0.5*dAlpha)/(2*dAlpha);
	} else {
		println("Specify correct distribution type");
		return 0;
	}
	if(s_iLinkFunction == LOG_LAMBDA){
		avScore[0] *= dLambda_t; // Apply chain rule
		amInvFisher[0] /= sqr(dLambda_t);	
	}
	return 1;
}

Scaling(const amScaling, const mInvFisher)
{
	if(s_iScalingChoice == INV_FISHER){
		amScaling[0] = mInvFisher;
		return 1;
	} else if(s_iScalingChoice == INV_SQRT_FISHER){
		amScaling[0] = sqrt(mInvFisher);
		return 1;
	} else {
		println("Specify correct scaling type");
	}
	return 0;
}

Init(const iMaxpq, const dAlpha, const dKappa)
{
	decl dlambda_t, vscore, minvfisher, dllik, mscale;
	for(decl p = 0; p < iMaxpq; p++)
	{
		if(s_iLinkFunction == LAMBDA)
			dlambda_t = s_vf[p];
		else
			dlambda_t = exp(s_vf[p]);
		SystemMatrices(s_mY[p], dlambda_t, dAlpha, dKappa, &vscore, &minvfisher, &dllik);
		s_vScore[p] = vscore;
		Scaling(&mscale, minvfisher);
		s_vScaledSc[p] = mscale*vscore;
	}
}

TransformPar(const vP, const adomega, const avA, const avB, const adalpha, const adkappa)
{
	if(s_iLinkFunction == LAMBDA)
		adomega[0] = exp(vP[0]);
	else
		adomega[0] = vP[0];
	avA[0] = reversec(vP[1:s_ip]);
	avB[0] = reversec(vP[1+s_ip:s_ip+s_iq]);
	if(s_iDistribution == EXP){
		adalpha[0] = 0;
		adkappa[0] = 0;
	} else if(s_iDistribution == WEI || s_iDistribution == GAM || s_iDistribution == LOG_NORM){	
		adalpha[0] = vP[s_cPar-1];
		adkappa[0] = 0;
	} else if(s_iDistribution == GEN_GAM){	
		adalpha[0] = vP[s_cPar-2];
		adkappa[0] = vP[s_cPar-1];
	}	
}

LogLikelihood(const vP, const adLlik, const avSco, const amHes)
{
	decl domega, vA, vB, dalpha, dkappa, imaxpq, dlambda_t, vscore, minvfisher, mscale, iret, dllik, dloglik = 0;	
	TransformPar(vP, &domega, &vA, &vB, &dalpha, &dkappa);
	imaxpq = max(s_ip, s_iq);
	s_vf[:imaxpq-1] = double(domega/(1 - sumc(vB))); // Unconditional mean of factor f
	Init(imaxpq, dalpha, dkappa); // Init s_t	
	for(decl t = imaxpq; t < s_cT; t++)
	{				
		s_vf[t] = domega + s_vScaledSc[t-s_ip:t-1]*vA + s_vf[t-s_iq:t-1]*vB;
		if(s_iLinkFunction == LAMBDA)
			dlambda_t = s_vf[t];
		else
			dlambda_t = exp(s_vf[t]);	
		iret = SystemMatrices(s_mY[t], dlambda_t, dalpha, dkappa, &vscore, &minvfisher, &dllik);
		if(s_iSandw == 1) s_vFunc[t-imaxpq] = dllik; // For sandwich standard errors		
		if(iret == 0) return 0;
		s_vScore[t] = vscore;
		dloglik += dllik;
		iret = Scaling(&mscale, minvfisher);
		if(iret == 0) return 0;
		s_vScaledSc[t] = mscale*vscore;
	}
	adLlik[0] = double(dloglik/s_cT);
	if(isnan(adLlik[0]) || sumc(vB) > 1) return 0;	
	return 1;   
}

ParStdErr(const vP)
{
    decl mhessian, invhess, mvar, iresult;
	iresult = Num2Derivative(LogLikelihood, vP, &mhessian);		
	if(!iresult)
    {	print("Covar() failed in numerical second derivatives\n");
        return zeros(vP);
    }
	invhess = invertgen(mhessian, 2); 
	if(s_iStdErr == HESS){
    	return sqrt(diagonal(-invhess)/s_cT)';
	} else if(s_iStdErr == SAND){
		decl dh, mcontr_grad, dloglik, vllik, vptemp, vllik_h, vscore_t, minnerp;	 
		s_iSandw = 1;
		s_vFunc = zeros(s_cT-max(s_iq, s_ip), 1);
		dh = 1e-07;
		mcontr_grad = zeros(s_cT-max(s_iq, s_ip), s_cPar);
		minnerp = zeros(s_cPar, s_cPar);
		LogLikelihood(vP, &dloglik, 0, 0);
		vllik = s_vFunc;
		for(decl i = 0; i < s_cPar; i++)
		{
			vptemp = vP;
			vptemp[i] += dh;
			LogLikelihood(vptemp, &dloglik, 0, 0);
			vllik_h = s_vFunc;
			mcontr_grad[][i] = (vllik_h-vllik)./dh; // Matrix of contributions to the gradient
		}
		minnerp = (mcontr_grad'*mcontr_grad)/s_cT;
		mvar = invhess*minnerp*invhess;
		s_iSandw = 0;
    	return sqrt(diagonal(mvar)/s_cT)';
	}
	println("Choose HESS or SAND to obtain standard errors");
}

MaxLik(vp, const aParNames)
{
    decl ir, dloglik, vpplot, vse;
    MaxControl(-1, 1, 1);
    ir = MaxBFGS(LogLikelihood, &vp, &dloglik, 0, TRUE);	
    println(MaxConvergenceMsg(ir), " using numerical derivatives");
	println("Log likelihood value = ", "%f", dloglik*s_cT);
	vse = ParStdErr(vp);
	vpplot = vp;
	if(s_iLinkFunction == LAMBDA){
		vse[0] = sqrt(sqr(vse[0])*exp(2*vp[0])); // Delta method
		vpplot[0] = exp(vp[0]);
	}
	print("Parameters with standard errors:", "%cf", {"%12.5g", " (%7.5f)"}, "%r", aParNames, vpplot ~ vse);
	if(s_iDistribution == GEN_GAM && vp[s_cPar-1] > 1000){
		println("The value of the shape parameter kappa is estimated at a large value (>1000), ");
		println("It is advised to apply the log-normal distribution to the dataset.");
		println("The Log-Normal distribution is a limiting case (kappa --> .Inf) of the Generalised Gamma distribution.");
		println("Choose s_iDistribution == LOG_NORM to select the Log-Normal distribution.");
	}	
	return vp;
}

PlotData()
{
	DrawXMatrix(0, s_mY, "Data", range(1, s_cT), "", 0, 3);
	if(s_iLinkFunction == LAMBDA)
		DrawXMatrix(0, s_vf, "Lambda_t", range(1, s_cT), "", 0, 2);
	else if(s_iLinkFunction == LOG_LAMBDA)
		DrawXMatrix(0, exp(s_vf), "Lambda_t", range(1, s_cT), "", 0, 2);
	DrawXMatrix(1, s_vScore, "Score", range(1, s_cT), "", 0, 3);
	DrawXMatrix(2, s_vScaledSc, "Scaled score", range(1, s_cT), "", 0, 3);	
}

CompareDuration(const domega, const vA, const vB, const dalpha, const dkappa, const ichoice1, const ichoice2)
{
	decl vp, vf1, vf2, aparnames;
	s_iDistribution = ichoice1;
	[vp, aparnames] = StartingValues(domega, vA, vB, dalpha, dkappa); 
	s_cPar = rows(vp);
	vp = MaxLik(vp, aparnames);
	vf1 = s_vf;
	s_iDistribution = ichoice2;
	[vp, aparnames] = StartingValues(domega, vA, vB, dalpha, dkappa);
	s_cPar = rows(vp);
	vp = MaxLik(vp, aparnames);
	vf2 = s_vf;
	DrawXMatrix(0, s_mY, "Data", range(1, s_cT), "", 0, 1);
	if(s_iLinkFunction == LAMBDA){
		DrawXMatrix(0, vf1, "Lambda_t, option 1", range(1, s_cT), "", 0, 2);
		DrawXMatrix(0, vf2, "Lambda_t, option 2", range(1, s_cT), "", 0, 3);
		DrawXMatrix(1, vf1, "Lambda_t, option 1", range(1, s_cT), "", 0, 2);
		DrawXMatrix(1, vf2, "Lambda_t, option 2", range(1, s_cT), "", 0, 3);		
	}	
	else if(s_iLinkFunction == LOG_LAMBDA){
		DrawXMatrix(0, exp(vf1), "Lambda_t, option 1", range(1, s_cT), "", 0, 2);
		DrawXMatrix(0, exp(vf2), "Lambda_t, option 2", range(1, s_cT), "", 0, 3);
		DrawXMatrix(1, exp(vf1), "Lambda_t, option 1", range(1, s_cT), "", 0, 2);
		DrawXMatrix(1, exp(vf2), "Lambda_t, option 2", range(1, s_cT), "", 0, 3);		
	}	
	ShowDrawWindow();	
}

main()
{
	// Univariate GAS model for modelling durations. See GAScodeDuration.pdf for explanation of the program
	decl dtime = 0, mdata, dscaling, dloglik, domega, vA, vB, vp, dalpha, dkappa, aparnames, ichoice1, ichoice2;

//*************************************************************************************************
//  START OF USER INPUT	
//	Load data										   
//	Specify scaling, distribution, link function, order of gas model,	   
//	choice of standard errors and starting values 					   
//*************************************************************************************************

	// Load data
	mdata = loadmat("3m9912-adur.xls"); s_mY = mdata[][0]';
	dscaling = 1; // Scaling data can improve stability, 1 for no scaling
	s_mY.*= dscaling;
	// Distribution: EXP, WEI, GAM, GEN_GAM, LOG_NORM
	s_iDistribution = EXP;
	// Link function: LAMBDA (f_t = lambda_t), LOG_LAMBDA (f_t = log(lambda_t))
	s_iLinkFunction = LAMBDA;
	// Scaling score: INV_FISHER, INV_SQRT_FISHER
	s_iScalingChoice = INV_FISHER;
	// Order of GAS model p, q
	s_ip = 1; s_iq = 1;
	// Standard erros: HESS, SAND
	s_iStdErr = HESS;	
	// Starting values (note the dimensions of s_ip and s_iq)
	domega = 0.05;  
	vA = <0.12>'; // Extend for higher orders of p, separate with comma's. 
	vB = <0.92>'; // Extend for higher orders of q, separate with comma's. B_1 + ... + B_q < 1
	dalpha = 0.18; // Shape parameter
	dkappa = 1; // Shape parameter
	
//*************************************************************************************************
//  END OF USER INPUT							      
//*************************************************************************************************

	s_cT = columns(s_mY);
	s_vf = s_vScore = s_vScaledSc = zeros(1, s_cT);
	[vp, aparnames] = StartingValues(domega, vA, vB, dalpha, dkappa); // Note that the number of elements in vp varies with the choice of distribution
	s_cPar = rows(vp);
	//LogLikelihood(vp, &dloglik, 0, 0);
	//println("Log Likelihood value = ", "%f", dloglik*s_cT);
	vp = MaxLik(vp, aparnames);
	PlotData();
	ShowDrawWindow();
	//ichoice1 = EXP;
	//ichoice2 = WEI;
	//CompareDuration(domega, vA, vB, dalpha, dkappa, ichoice1, ichoice2);
	println("Time: ", timespan(dtime));	
}
